# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/05-preprocessor.ipynb (unless otherwise specified).

__all__ = ['Preprocessor', 'preprocess_description', 'run']

# Cell

import pandas as pd
import fastcore.basics as fcb
from .core import logger
from pymystem3 import Mystem
import re

# Cell

class Preprocessor:
    def __init__(self):

        # get mystem object:
        self.mystem = Mystem()

# Cell

@fcb.patch_to(Preprocessor)

def preprocess_description(self, vacancy_df):

    """
    FUNC TO PREPROCESS VACANCIES (CLEAN PUNCTUATION, CLEAN DIGITS AND LEMMATIZE)

    RETURN: pd.DataFrame

    """

    # lower text:
    vacancy_df['description'] = vacancy_df['description'].str.lower()

    # replace ",", ".", ":", ";", "-" with gaps:
    vacancy_df['description'] = vacancy_df['description'].apply(lambda x: re.sub(r'[,.:;-]', ' ', x))

    # replace "c++" and "c#" in order not to be dropped:
    vacancy_df['description'] = vacancy_df['description'].apply(lambda x: re.sub(r'(c\+\+|с\+\+)', 'cplusplus ', x))
    vacancy_df['description'] = vacancy_df['description'].apply(lambda x: re.sub(r'(c#|с#)', 'csharp ', x))

    # remove all punctuation:
    vacancy_df['description'] = vacancy_df['description'].apply(lambda x: re.sub(r'[^\w\s]', '', x))

    # remove digits if surrounded by spaces:
    vacancy_df['description'] = vacancy_df['description'].apply(lambda x: re.sub(r'(?<=\s)\d+(?=\s)', '', x))

    # lemmatize:
    def lemmatize(text):
        mystem = self.mystem
        tokens = mystem.lemmatize(text)
        tokens = [token for token in tokens if token != " " and token != "\n"]
        text = " ".join(tokens)
        return text

    vacancy_df['description'] = vacancy_df['description'].apply(lambda x: lemmatize(x))

    # remove duplicated gaps:
    vacancy_df['description'] = vacancy_df['description'].apply(lambda x: ' '.join(x.split()))

    return vacancy_df

# Cell

@fcb.patch_to(Preprocessor)

def run(self, vacancy_df):

    """
    FUNC TO RUN PREPROCESSING

    RETURN: pd.DataFrame

    """

    logger.info('SUCCESS. started preprocessing vacancies')


    # get vacancy_df:
    vacancy_df['description'] = vacancy_df['description'] + vacancy_df['name']

    # preprocess descriptions:
    vacancy_df = self.preprocess_description(vacancy_df)

    # filter currency:
    vacancy_df = vacancy_df[vacancy_df.salary_currency.isin([None, 'RUR'])]

    logger.info('SUCCESS. finished preprocessing vacancies')

    return vacancy_df


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01-postgres.ipynb (unless otherwise specified).

__all__ = ['PostgresConnector', 'execute', 'create_schema', 'drop_schema', 'check_connection', 'create_table',
           'download', 'table_exists', 'delete_table', 'save', 'get_schemas_list']

# Cell
from sqlalchemy import create_engine, inspect
import pandas as pd
from sshtunnel import SSHTunnelForwarder
import numpy as np
import fastcore.basics as fcb
from .core import logger

# Cell
class PostgresConnector:
    def __init__(
        self,
        ssh_host,
        ssh_username,
        ssh_password,
        local_host,
        postgres_username,
        postgres_password,
        database,
    ):
        self.ssh_host = ssh_host
        self.ssh_username = ssh_username
        self.ssh_password = ssh_password
        self.local_host = local_host
        self.postgres_username = postgres_username
        self.postgres_password = postgres_password
        self.database = database

        self.server = SSHTunnelForwarder(
            (ssh_host, 22),
            ssh_username=ssh_username,
            ssh_password=ssh_password,
            remote_bind_address=(local_host, 5432)
        )

        logger.info(f"SUCCESS: initialized PostgresConnector")

    def create_engine(self):
        server = self.server
        local_port = str(server.local_bind_port)
        engine = create_engine(f'postgresql://{self.postgres_username}:{self.postgres_password}@{self.local_host}:{local_port}/{self.database}')
        return engine

# Cell
@fcb.patch_to(PostgresConnector)
def execute(self, sql_command):

    server = self.server

    server.start()
    engine = self.create_engine()
    engine.execute(sql_command)
    server.stop()

    logger.info(f"Executed command: {sql_command} in database: {self.database}")

# Cell
@fcb.patch_to(PostgresConnector)
def create_schema(self, schema_name):
    self.execute(f'CREATE SCHEMA {schema_name};')
    logger.info(f"SUCCESS: schema {schema_name} was created in database: {self.database}")

# Cell
@fcb.patch_to(PostgresConnector)
def drop_schema(self, schema_name):
    self.execute(f'DROP SCHEMA {schema_name} CASCADE;')
    logger.info(f"SUCCESS: schema {schema_name} was dropped from database: {self.database}")

# Cell
@fcb.patch_to(PostgresConnector)
def check_connection(self):
    try:
        server = self.server
        server.start()
        engine = self.create_engine()
        n = pd.read_sql("SELECT COUNT(*) FROM pg_catalog.pg_type", engine).values[0][0]
        server.stop()
        assert (n > 100), f"pg_catalog.pg_type should have > 400 rows, got {n} instead"
        logger.info("Connection to database established")
    except Exception as e:
        logger.error(f"Connection to database has failed")
        raise e

# Cell
@fcb.patch_to(PostgresConnector)
def create_table(self,
    df,
    table_name: str,
    schema: str = "public",
):

    ## Run some checks:
    # column "inserted_into_db_msk" will be created automatically, it shouldn't be passed:
    if ("inserted_into_db_msk" in df.columns):  #
        raise NotImplementedError("Did not expect 'inserted_into_db_msk' column in df")

    # check that table name starts not from numeric:
    if not table_name[0].isalpha():
        raise Exception(f"Table name shouldn't start from numeric value. Passed table name: {table_name}")


    # numpy to Postgres type map
    np_to_postgres_type = {
        np.dtype("int64"): "bigint",
        np.dtype("float32"): "real",
        np.dtype("float64"): "real",
        np.dtype("O"): "text",
        np.dtype("datetime64[ns]"): "timestamp",
        np.dtype("bool"): "boolean"}

    # part of query for defining columns names and types, according to the types in initial df:
    sql_columns_list = []

    for column, col_type in zip(df.columns, df.dtypes):
        sql_columns_list.append(f"{column} {np_to_postgres_type[col_type]}")

    # creating final command
    sql_column_creation_command = ", ".join(sql_columns_list)

    final_sql_command = f"""
                         CREATE TABLE {schema}.{table_name}
                         ({sql_column_creation_command},
                         inserted_into_db_msk timestamp DEFAULT (current_timestamp AT TIME ZONE 'Europe/Moscow'));
                         """

    # Finally creating that command
    server = self.server
    server.start()
    engine = self.create_engine()
    engine.execute(final_sql_command)
    server.stop()

    logger.info(f"Table named {table_name} was created")

# Cell
@fcb.patch_to(PostgresConnector)
def download(self, sql_query: str):

    server = self.server
    server.start()
    engine = self.create_engine()
    result = pd.read_sql_query(sql=sql_query, con=engine)
    server.stop()
    return result

# Cell
@fcb.patch_to(PostgresConnector)
def table_exists(self, table_name: str, schema: str):
    try:
        sql_query_to_check_existence = f"SELECT * FROM {schema}.{table_name} LIMIT 1"
        self.download(sql_query=sql_query_to_check_existence)
        return True
    except:
        return False

# Cell
@fcb.patch_to(PostgresConnector)
def delete_table(
    self,
    table_name: str,
    schema: str,
):
    if not self.table_exists(table_name=table_name, schema=schema):
        raise Exception(
            f"Trying to delete non-existent table. Table {schema}.{table_name} doesn't exist"
        )


    server = self.server
    server.start()
    engine = self.create_engine()

    engine.execute(f"DROP TABLE IF EXISTS {schema}.{table_name};")
    server.stop()
    logger.info(f"Table named {table_name} was deleted")

# Cell
@fcb.patch_to(PostgresConnector)
def save(
    self,
    df: pd.DataFrame,
    table_name: str,
    schema: str,
    if_table_exists: str = "append",
):
    # check if correct parameters are inserted:
    if df.empty:
        raise Exception("The dataframe is empty")
    if if_table_exists not in ["append", "replace"]:
        raise ValueError(f"'{if_table_exists}' is not valid for if_table_exists")

    # if table doesn't exist - create it:
    if not self.table_exists(table_name=table_name, schema=schema):
        logger.info(f"The table named {table_name} doesn't exist. It will be created")
        self.create_table(df=df, table_name=table_name, schema=schema)

    if if_table_exists == "replace":
        logger.info("An existing table will be dropped and filled with fresh values")
        self.delete_table(table_name=table_name, schema=schema)
        self.save(df=df, table_name=table_name, schema=schema, if_table_exists="append")

    if if_table_exists == "append":
        logger.info(f"Filling table {table_name} with values")
        server = self.server
        server.start()
        engine = self.create_engine()
        df.to_sql(name=table_name, con=engine, schema=schema, index=False, if_exists="append")
        server.stop()

# Cell
@fcb.patch_to(PostgresConnector)
def get_schemas_list(self):
    return self.download('SELECT nspname FROM pg_namespace').rename(columns={'nspname':'schema'})


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/04-hh_parser.ipynb (unless otherwise specified).

__all__ = ['HeadHunterScrapper', 'get_urls_from_professional_role', 'get_url_vs_responses', 'get_vacancies_info']

# Cell

import pandas as pd
from random import choice, uniform
import requests
import numpy as np
import fastcore.basics as fcb
from .core import logger
from .postgres import PostgresConnector
import time
from tqdm import tqdm
import re

# Cell

class HeadHunterScrapper:
    def __init__(self, base_url='https://api.hh.ru/vacancies'):

        # self.proxies_list = proxies_list
        self.base_url = base_url

# Cell

@fcb.patch_to(HeadHunterScrapper)

def get_urls_from_professional_role(self, professional_role, per_page=100, pages=20, sleep=(1,3)):

    logger.info(f'SUCCESS: started scrapping vacancies URLs for professional_role = {professional_role}')

    # get API's base url:
    base_url = self.base_url

    # get vacancies' urls of entered professional_role:
    urls = []
    responses = []
    ids = []
    for page in range(pages):

        logger.info(f'SUCCESS: started scrapping page {page}')

        # make request:
        params = {'professional_role':professional_role, 'area':1, 'per_page':per_page, 'page':page, 'responses_count_enabled':True} # Moscow
        # json = requests.get(base_url, params=params, proxies=proxies).json()
        json = requests.get(base_url, params=params).json()

        # break if there're no more vacancies of professional_role:
        if len(json['items'])==0:
            break

        # get list of vacancies' urls and append it:
        vacancy_urls      = pd.Series(json['items']).apply(lambda x: x['url']).to_list()
        vacancy_responses = pd.Series(json['items']).apply(lambda x: x['counters']['responses']).to_list() # кол-во откликов
        vacancy_ids       = pd.Series(json['items']).apply(lambda x: x['id']).to_list()

        urls      += vacancy_urls
        responses += vacancy_responses
        ids       += vacancy_ids

        # sleep:
        min_sleep = sleep[0]
        max_sleep = sleep[1]
        time.sleep(uniform(min_sleep,max_sleep))

        logger.info(f'SUCCESS: finished scrapping page {page}')

    # create dataframe:
    vacancy_url_df = pd.DataFrame({'professional_role':professional_role, 'url':urls, 'responses':responses, 'id':ids})

    logger.info(f'SUCCESS: finished scrapping vacancies URLs for professional_role = {professional_role}')

    return vacancy_url_df

# Cell

@fcb.patch_to(HeadHunterScrapper)

def get_url_vs_responses(self, professional_roles, per_page=100, pages=20, sleep=(1,3), postgres_creds=None, save_to_postgres=False):

    logger.info(f'SUCCESS: started scrapping vacancies URLs for professional_roles = {professional_roles}')

    dfs = []
    for professional_role in tqdm(professional_roles):

        # try up to 5 times:
        for attempt in range(5):
            try:
                df = self.get_urls_from_professional_role(professional_role, per_page=per_page, pages=pages, sleep=sleep)
                dfs.append(df)
                break

            except requests.exceptions.ConnectionError:
                time.sleep(10)
                logger.info('ERROR: caught ConnectionError. Retrying.')

    vacancy_url_df = pd.concat(dfs)[['url', 'responses']]

    logger.info(f'SUCCESS: finished scrapping vacancies URLs for professional_roles = {professional_roles}')

    if save_to_postgres:
        logger.info(f'SUCCESS: started saving to Postgres')
        pc = PostgresConnector(**postgres_creds)
        pc.save(df=vacancy_url_df, table_name='url_vs_responses', schema='main_schema', if_table_exists='replace')
        logger.info(f'SUCCESS: finished saving to Postgres')

    return vacancy_url_df

# Cell

@fcb.patch_to(HeadHunterScrapper)

def get_vacancies_info(self, urls, sleep=(1,3), postgres_creds=None, save_to_postgres=False):

    logger.info(f'SUCCESS: started scrapping vacancies info')

    dfs = []
    for url in tqdm(urls):

        # try up to 5 times:
        for attempt in range(5):
            try:
                logger.info(f'SUCCESS: started scrapping {url}')
                json = requests.get(url).json()

                df = pd.DataFrame({'id':[json['id']],
                                   'name':[json['name']],
                                   'salary':[json['salary']],
                                   'key_skills':[json['key_skills']],
                                   'experience':[json['experience']],
                                   'description':[json['description']],
                                   'schedule':[json['schedule']]})

                logger.info(f'SUCCESS: finished scrapping {url}')
                break

            except requests.exceptions.ConnectionError:
                time.sleep(10)
                logger.info('ERROR: caught ConnectionError. Retrying.')

        dfs.append(df)

        # sleep:
        min_sleep = sleep[0]
        max_sleep = sleep[1]
        time.sleep(uniform(min_sleep,max_sleep))

    logger.info(f'SUCCESS: finished scrapping vacancies info')

    # preproccess data:
    logger.info(f'SUCCESS: started preprocessing data')
    vacancy_info_df = pd.concat(dfs)

    # edit description:
    vacancy_info_df.description = vacancy_info_df.description.apply(lambda x: re.sub(re.compile('<.*?>'), '', x))

    # edit key_skills:
    vacancy_info_df.key_skills = vacancy_info_df.key_skills.apply(lambda x: ', '.join([d['name'] for d in x]))

    # edit experience:
    vacancy_info_df.experience = vacancy_info_df.experience.apply(lambda x: x['name'])

    # edit schedule:
    vacancy_info_df.schedule = vacancy_info_df.schedule.apply(lambda x: x['id'])

    # edit salary:
    vacancy_info_df['salary_from'] = vacancy_info_df.salary.apply(lambda x: x['from'] if type(x)==dict else None)
    vacancy_info_df['salary_to'] = vacancy_info_df.salary.apply(lambda x: x['to'] if type(x)==dict else None)
    vacancy_info_df['salary_currency'] = vacancy_info_df.salary.apply(lambda x: x['currency'] if type(x)==dict else None)

    # choose columns:
    vacancy_info_df = vacancy_info_df[['id', 'name', 'key_skills', 'experience', 'description', 'schedule', 'salary_from', 'salary_to', 'salary_currency']]

    logger.info(f'SUCCESS: finished preprocessing data')

    if save_to_postgres:
        logger.info(f'SUCCESS: started saving to Postgres')
        pc = PostgresConnector(**postgres_creds)
        pc.save(df=vacancy_info_df, table_name='vacancy_info', schema='main_schema', if_table_exists='replace')
        logger.info(f'SUCCESS: finished saving to Postgres')

    return vacancy_info_df